// Copyright 2019 Google LLC
// Copyright 2019 The libgav1 Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// -----------------------------------------------------------------------------

// This file is just a convenience to separate out all the CDF constant
// definitions from the symbol decoder context functions.

// The first half of this file stores the original AOM CDFs.
// You can #define PRINT_AOM_CDF_CONVERTED_TO_WP2 in this file and
// compile this code to get an executable that prints the AOM CDFs under the
// WP2 conventions.
// Compile doing so if your build folder is at the source root.
// clang++ -x c++ -L. -I ../ -o main ../src/common/lossy/aom/cdfs.inc -lwp2
// TODO(vrabaud) simplify before final shipping.

#ifdef PRINT_AOM_CDF_CONVERTED_TO_WP2
#include <cassert>
#include <iostream>
#include <string>

#include "src/utils/ans.h"
#endif

namespace WP2 {

// Maximum number of coordinates in an AOM CDF array (excluding the one for the
// coefficients).
static constexpr uint32_t kCoordMaxSize = 4;

// The following struct/functions convert a cluster id as defined in SymbolInfo
// into a set of coodinates for the AOM C arrays containing the CDFs.

// Helper struct for GetCoordinates, that reads the extent one at a time.
template <uint32_t U>
struct CoordinateFiller {
  template <class T>
  static void Do(const T& cdf_array, uint32_t* const in,
                 uint32_t* const num_extents) {
    if (U < std::rank<T>::value) {
      in[U] = std::extent<T, U>::value;
    }
    CoordinateFiller<U - 1>::Do(cdf_array, in, num_extents);
  }
};

template <>
struct CoordinateFiller<0> {
  template <class T>
  static void Do(const T& cdf_array, uint32_t* const in,
                 uint32_t* const num_extents) {
    if (num_extents != nullptr) {
      *num_extents = std::rank<T>::value;
    }
    in[0] = std::extent<T, 0>::value;
  }
};

// Converts a SymbolInfo cluster for a CDF as a set of coordinates in the
// correponding AOM CDF array (which have different number of coordinates, hence
// the template).
// E.g., if we have a uint16_t kCDFs[3][7][10], we convert 15 into (2,1) because
// cluster 15 = 2 * 7 + 1 represents the CDF: kCDFs[2][1][].
template <class U>
std::array<uint32_t, std::rank<U>::value - 1> GetCoordinates(
                                         uint32_t cluster, const U& cdf_array) {
  constexpr uint32_t size = std::rank<U>::value - 1;
  assert(size <= kCoordMaxSize);
  uint32_t extents[kCoordMaxSize];
  CoordinateFiller<size - 1>::Do(cdf_array, extents, nullptr);
  uint32_t prod = 1;
  for (uint32_t i = 1; i < size; ++i) prod *= extents[i];
  std::array<uint32_t, size> res;
  for (uint32_t i = 0; i < size - 1; ++i) {
    res[i] = cluster / prod;
    cluster %= prod;
    prod /= extents[i + 1];
  }
  res[size - 1] = cluster;
  return res;
}

// Helper functions for the GetCDF function.
template<class U>
static const uint16_t* GetCDFHelper(const std::array<uint32_t, 4>& c,
                                    const U& cdf_array) {
  static_assert(std::rank<U>::value == 5, "wrong cdf array size");
  return cdf_array[c[0]][c[1]][c[2]][c[3]];
}

template<class U>
static const uint16_t* GetCDFHelper(const std::array<uint32_t, 3>& c,
                                    const U& cdf_array) {
  static_assert(std::rank<U>::value == 4, "wrong cdf array size");
  return cdf_array[c[0]][c[1]][c[2]];
}

template<class U>
static const uint16_t* GetCDFHelper(const std::array<uint32_t, 2>& c,
                                    const U& cdf_array) {
  static_assert(std::rank<U>::value == 3, "wrong cdf array size");
  return cdf_array[c[0]][c[1]];
}

template<class U>
static const uint16_t* GetCDFHelper(const std::array<uint32_t, 1>& c,
                                    const U& cdf_array) {
  static_assert(std::rank<U>::value == 2, "wrong cdf array size");
  return cdf_array[c[0]];
}

// For a given AOM cdf array 'cdf_array' and a given cluster as defined in
// SymbolInfo, returns the corresponding CDF.
template <class U>
const uint16_t* GetCDF(uint32_t cluster, const U& cdf_array) {
  constexpr uint32_t size = std::rank<U>::value - 1;
  const std::array<uint32_t, size> c = GetCoordinates(cluster, cdf_array);
  return GetCDFHelper(c, cdf_array);
}

namespace libgav1 {

#ifdef PRINT_AOM_CDF_CONVERTED_TO_WP2

enum PlaneType : uint8_t { kPlaneTypeY, kPlaneTypeUV, kNumPlaneTypes };
enum {
  kBooleanFieldCdfSize = 3,
  kAllZeroContexts = 13,
  kEobPtContexts = 2,
  kEobPt4SymbolCount = 3,
  kEobPt8SymbolCount = 4,
  kEobPt16SymbolCount = 5,
  kEobPt32SymbolCount = 6,
  kEobPt64SymbolCount = 7,
  kEobPt128SymbolCount = 8,
  kEobPt256SymbolCount = 9,
  kEobPt512SymbolCount = 10,
  kEobPt1024SymbolCount = 11,
  kEobExtraContexts = 9,
  kCoeffBaseEobContexts = 4,
  kCoeffBaseEobSymbolCount = 3,
  kCoeffBaseContexts = 42,
  kCoeffBaseSymbolCount = 4,
  kCoeffBaseRangeContexts = 21,
  kCoeffBaseRangeSymbolCount = 3
};  // anonymous enum

constexpr uint16_t kDefaultEobPt4Cdf[kNumPlaneTypes][kEobPtContexts]
                                    [kEobPt4SymbolCount + 1] = {
                                         {{{30788, 27873, 0, 0},
                                           {30885, 28297, 0, 0}},
                                          {{23080, 18205, 0, 0},
                                           {25005, 18121, 0, 0}}},
                                         {{{27603, 23822, 0, 0},
                                           {30909, 26429, 0, 0}},
                                          {{18509, 13660, 0, 0},
                                           {23902, 15775, 0, 0}}},
                                         {{{23887, 17800, 0, 0},
                                           {30122, 22712, 0, 0}},
                                          {{14527, 9202, 0, 0},
                                           {22471, 13013, 0, 0}}},
                                         {{{18022, 10635, 0, 0},
                                           {27985, 17358, 0, 0}},
                                          {{6724, 3059, 0, 0},
                                           {13495, 4574, 0, 0}}}};
constexpr uint16_t kDefaultEobPt8Cdf[kNumPlaneTypes][kEobPtContexts]
                                    [kEobPt8SymbolCount + 1] = {
    {{31729, 30788, 27873, 0, 0}, {32097, 30885, 28297, 0, 0}},
    {{27818, 23080, 18205, 0, 0}, {29414, 25005, 18121, 0, 0}}
};

constexpr uint16_t kDefaultEobPt16Cdf[kNumPlaneTypes][kEobPtContexts]
                                     [kEobPt16SymbolCount + 1] = {
    {{31928, 31729, 30788, 27873, 0, 0}, {32398, 32097, 30885, 28297, 0, 0}},
    {{29521, 27818, 23080, 18205, 0, 0}, {30864, 29414, 25005, 18121, 0, 0}}
};
constexpr uint16_t kDefaultEobPt32Cdf[kNumPlaneTypes]
                      [kEobPtContexts][kEobPt32SymbolCount + 1] = {
  {{32368, 32248, 31791, 30666, 26226, 0, 0},
   {32558, 32363, 31453, 29442, 25231, 0, 0}},
  {{30132, 28495, 25180, 20974, 12367, 0, 0},
   {30982, 29589, 25866, 21411, 13714, 0, 0}}},
};
constexpr uint16_t kDefaultEobPt64Cdf[kNumPlaneTypes]
                      [kEobPtContexts][kEobPt64SymbolCount + 1] = {
    {{32439, 32270, 31667, 30984, 29503, 25010, 0, 0},
     {32433, 32038, 31309, 27274, 24013, 19771, 0, 0}},
    {{29263, 27464, 22682, 18954, 15084, 9398, 0, 0},
     {31205, 30068, 27892, 21857, 18062, 10288, 0, 0}}
};
constexpr uint16_t kDefaultEobPt128Cdf[kNumPlaneTypes][kEobPtContexts]
                                      [kEobPt128SymbolCount + 1] = {
    {{32549, 32286, 31628, 30677, 29088, 26740, 20182, 0, 0},
     {32397, 32069, 31514, 27938, 23289, 20206, 15271, 0, 0}},
    {{27523, 25312, 19888, 16916, 12735, 8836, 5160, 0, 0},
     {30714, 29296, 26899, 18536, 14526, 12178, 6016, 0, 0}}
};
constexpr uint16_t kDefaultEobPt256Cdf[kNumPlaneTypes][kEobPtContexts]
                                      [kEobPt256SymbolCount + 1] = {
  {{32458, 32184, 30881, 29179, 26600, 24157, 21416, 17116, 0, 0},
   {31770, 30918, 29770, 27164, 15427, 12880, 9869, 7185, 0, 0}},
  {{30248, 29528, 26816, 23898, 20191, 15210, 12814, 8600, 0, 0},
   {30565, 28638, 25333, 22029, 12116, 9087, 7159, 5507, 0, 0}}
};
constexpr uint16_t kDefaultEobPt512Cdf
    [kNumPlaneTypes][kEobPt512SymbolCount + 1] =
  {32127, 31785, 29061, 27338, 22534, 17810, 13980, 9356, 6707, 0, 0},
  {27673, 26322, 22772, 19414, 16751, 14782, 11849, 6639, 3628, 0, 0}
};

constexpr uint16_t kDefaultEobPt1024Cdf[kNumPlaneTypes]
                                       [kEobPt1024SymbolCount + 1] = {
  {32375, 32347, 32017, 31145, 29608, 26416, 19423, 14721, 10197, 6938, 0, 0},
  {30903, 30780, 29838, 28526, 22235, 16230, 11414, 5513, 4222, 984, 0, 0}
};

/* clang-format off */
constexpr uint16_t kDefaultEobExtraCdf[kNumPlaneTypes][kEobExtraContexts]
                                      [kBooleanFieldCdfSize] = {
  {
    {
      {{15807, 0, 0}, {15545, 0, 0}, {25147, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}},
      {{13699, 0, 0}, {10243, 0, 0}, {19391, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    },
    {
      {{12367, 0, 0}, {15743, 0, 0}, {19923, 0, 0}, {19895, 0, 0},
       {18674, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}},
      {{12087, 0, 0}, {12067, 0, 0}, {17518, 0, 0}, {17751, 0, 0},
       {17840, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    },
    {
      {{8863, 0, 0}, {15574, 0, 0}, {16598, 0, 0}, {15073, 0, 0},
       {18942, 0, 0}, {16958, 0, 0}, {20732, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}},
      {{8809, 0, 0}, {11969, 0, 0}, {13747, 0, 0}, {16565, 0, 0},
       {14882, 0, 0}, {18624, 0, 0}, {20758, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    },
    {
      {{5369, 0, 0}, {16441, 0, 0}, {14697, 0, 0}, {13184, 0, 0},
       {12047, 0, 0}, {14336, 0, 0}, {13208, 0, 0}, {22618, 0, 0},
       {23963, 0, 0}},
      {{7836, 0, 0}, {11935, 0, 0}, {20741, 0, 0}, {16098, 0, 0},
       {12854, 0, 0}, {17662, 0, 0}, {15106, 0, 0}, {18985, 0, 0},
       {4012, 0, 0}}
    },
    {
      {{9362, 0, 0}, {10923, 0, 0}, {14336, 0, 0}, {16384, 0, 0},
       {15672, 0, 0}, {20207, 0, 0}, {15448, 0, 0}, {10373, 0, 0},
       {11398, 0, 0}},
      {{16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    }
  },
  {
    {
      {{15297, 0, 0}, {12545, 0, 0}, {21411, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}},
      {{12433, 0, 0}, {11101, 0, 0}, {17950, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    },
    {
      {{12338, 0, 0}, {12106, 0, 0}, {17401, 0, 0}, {15798, 0, 0},
       {18111, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}},
      {{10651, 0, 0}, {10740, 0, 0}, {14118, 0, 0}, {16726, 0, 0},
       {16883, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    },
    {
      {{10359, 0, 0}, {11756, 0, 0}, {17118, 0, 0}, {15373, 0, 0},
       {17299, 0, 0}, {12563, 0, 0}, {13257, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}},
      {{8548, 0, 0}, {10288, 0, 0}, {15031, 0, 0}, {13852, 0, 0},
       {13500, 0, 0}, {14356, 0, 0}, {13924, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    },
    {
      {{6777, 0, 0}, {12454, 0, 0}, {15037, 0, 0}, {13090, 0, 0},
       {14119, 0, 0}, {15461, 0, 0}, {10970, 0, 0}, {15219, 0, 0},
       {17138, 0, 0}},
      {{6183, 0, 0}, {11299, 0, 0}, {12336, 0, 0}, {15033, 0, 0},
       {13488, 0, 0}, {17533, 0, 0}, {12471, 0, 0}, {10297, 0, 0},
       {3771, 0, 0}}
    },
    {
      {{6163, 0, 0}, {21464, 0, 0}, {16042, 0, 0}, {16208, 0, 0},
       {11902, 0, 0}, {9244, 0, 0}, {12890, 0, 0}, {19299, 0, 0},
       {9684, 0, 0}},
      {{16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    }
  },
  {
    {
      {{13785, 0, 0}, {12256, 0, 0}, {17883, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}},
      {{12678, 0, 0}, {13324, 0, 0}, {15482, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    },
    {
      {{13629, 0, 0}, {11281, 0, 0}, {13809, 0, 0}, {11858, 0, 0},
       {13679, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}},
      {{12232, 0, 0}, {12104, 0, 0}, {12143, 0, 0}, {13645, 0, 0},
       {17906, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    },
    {
      {{12935, 0, 0}, {11266, 0, 0}, {15283, 0, 0}, {12501, 0, 0},
       {14415, 0, 0}, {9439, 0, 0}, {11290, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}},
      {{10727, 0, 0}, {9334, 0, 0}, {12767, 0, 0}, {12214, 0, 0},
       {11817, 0, 0}, {12623, 0, 0}, {17206, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    },
    {
      {{9456, 0, 0}, {11161, 0, 0}, {16242, 0, 0}, {13811, 0, 0},
       {14734, 0, 0}, {13834, 0, 0}, {8521, 0, 0}, {15847, 0, 0},
       {15688, 0, 0}},
      {{6189, 0, 0}, {7858, 0, 0}, {14131, 0, 0}, {12968, 0, 0},
       {12380, 0, 0}, {22881, 0, 0}, {17126, 0, 0}, {2570, 0, 0},
       {8047, 0, 0}}
    },
    {
      {{5770, 0, 0}, {16031, 0, 0}, {14930, 0, 0}, {13846, 0, 0},
       {13253, 0, 0}, {14132, 0, 0}, {15435, 0, 0}, {16992, 0, 0},
       {10110, 0, 0}},
      {{16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    }
  },
  {
    {
      {{12591, 0, 0}, {11979, 0, 0}, {12506, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}},
      {{11352, 0, 0}, {11913, 0, 0}, {9358, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    },
    {
      {{12530, 0, 0}, {11711, 0, 0}, {13609, 0, 0}, {10431, 0, 0},
       {12609, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}},
      {{12643, 0, 0}, {12209, 0, 0}, {11061, 0, 0}, {10472, 0, 0},
       {15435, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    },
    {
      {{12827, 0, 0}, {12241, 0, 0}, {11298, 0, 0}, {10281, 0, 0},
       {13210, 0, 0}, {10414, 0, 0}, {12437, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}},
      {{10016, 0, 0}, {7762, 0, 0}, {10693, 0, 0}, {11192, 0, 0},
       {15028, 0, 0}, {11078, 0, 0}, {13557, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    },
    {
      {{11326, 0, 0}, {10410, 0, 0}, {14265, 0, 0}, {12477, 0, 0},
       {12823, 0, 0}, {11474, 0, 0}, {11590, 0, 0}, {13368, 0, 0},
       {22212, 0, 0}},
      {{8120, 0, 0}, {7819, 0, 0}, {12060, 0, 0}, {8863, 0, 0},
       {12267, 0, 0}, {23210, 0, 0}, {23345, 0, 0}, {2403, 0, 0},
       {13515, 0, 0}}
    },
    {
      {{6704, 0, 0}, {10670, 0, 0}, {13155, 0, 0}, {12243, 0, 0},
       {15173, 0, 0}, {16150, 0, 0}, {12271, 0, 0}, {13779, 0, 0},
       {17255, 0, 0}},
      {{16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0}, {16384, 0, 0},
       {16384, 0, 0}}
    }
  }
};

constexpr uint16_t kDefaultCoeffBaseEobCdf[kNumPlaneTypes]
    [kCoeffBaseEobContexts] [kCoeffBaseEobSymbolCount + 1] = {
    {
      {{14931, 3713, 0, 0}, {3168, 1322, 0, 0}, {1924, 890, 0, 0},
       {7842, 3820, 0, 0}},
      {{11403, 2742, 0, 0}, {2256, 345, 0, 0}, {1110, 147, 0, 0},
       {3138, 887, 0, 0}}
    },
    {
      {{27051, 6291, 0, 0}, {2277, 1065, 0, 0}, {1218, 610, 0, 0},
       {3120, 1277, 0, 0}},
      {{20160, 4948, 0, 0}, {2088, 543, 0, 0}, {1959, 433, 0, 0},
       {1469, 345, 0, 0}}
    },
    {
      {{30982, 20156, 0, 0}, {2105, 1143, 0, 0}, {429, 300, 0, 0},
       {1620, 935, 0, 0}},
      {{13911, 8903, 0, 0}, {1340, 340, 0, 0}, {1024, 395, 0, 0},
       {993, 242, 0, 0}}
    },
    {
      {{30981, 30236, 0, 0}, {1936, 1106, 0, 0}, {944, 86, 0, 0},
       {635, 199, 0, 0}},
      {{19017, 10533, 0, 0}, {679, 359, 0, 0}, {5684, 4848, 0, 0},
       {3477, 174, 0, 0}}
    },
    {
      {{31043, 29319, 0, 0}, {1666, 833, 0, 0}, {311, 155, 0, 0},
       {356, 119, 0, 0}},
      {{21845, 10923, 0, 0}, {21845, 10923, 0, 0}, {21845, 10923, 0, 0},
       {21845, 10923, 0, 0}}
    }
};
/* clang-format on */

constexpr uint16_t kDefaultCoeffBaseCdf
    [kNumPlaneTypes]
    [kCoeffBaseContexts][kCoeffBaseSymbolCount + 1] = {
        {{{{28734, 23838, 20041, 0, 0}, {14686, 3027, 891, 0, 0},
           {20172, 6644, 2275, 0, 0},   {23322, 11650, 5763, 0, 0},
           {26460, 17627, 11489, 0, 0}, {30305, 26411, 22985, 0, 0},
           {12101, 2222, 839, 0, 0},    {19725, 6645, 2634, 0, 0},
           {24617, 14011, 7990, 0, 0},  {27513, 19929, 14136, 0, 0},
           {29948, 25562, 21607, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {17032, 5215, 2164, 0, 0},
           {21558, 8974, 3981, 0, 0},   {26821, 18894, 13067, 0, 0},
           {28553, 23445, 18877, 0, 0}, {29935, 26306, 22709, 0, 0},
           {13163, 2375, 1186, 0, 0},   {19245, 6516, 2520, 0, 0},
           {24322, 14146, 8256, 0, 0},  {28950, 22425, 16794, 0, 0},
           {31287, 28651, 25972, 0, 0}, {10119, 1466, 578, 0, 0},
           {17939, 5641, 2319, 0, 0},   {24455, 15066, 9464, 0, 0},
           {29746, 24467, 19982, 0, 0}, {31232, 28356, 25584, 0, 0},
           {10414, 2994, 1396, 0, 0},   {18045, 7296, 3554, 0, 0},
           {26095, 19023, 14106, 0, 0}, {30700, 27002, 23446, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}},
          {{26466, 16324, 11007, 0, 0}, {9728, 1230, 293, 0, 0},
           {17572, 4316, 1272, 0, 0},   {22748, 9822, 4254, 0, 0},
           {26235, 15906, 9267, 0, 0},  {29230, 22952, 17692, 0, 0},
           {8324, 893, 243, 0, 0},      {16887, 3844, 1133, 0, 0},
           {22846, 9895, 4302, 0, 0},   {26241, 15802, 9077, 0, 0},
           {28654, 21465, 15548, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {12567, 1998, 559, 0, 0},
           {18014, 4697, 1510, 0, 0},   {24390, 12582, 6251, 0, 0},
           {26852, 17469, 10790, 0, 0}, {28500, 21185, 14867, 0, 0},
           {8407, 743, 187, 0, 0},      {14095, 2663, 825, 0, 0},
           {22572, 10524, 5192, 0, 0},  {27273, 18419, 12351, 0, 0},
           {30092, 25353, 21270, 0, 0}, {8090, 810, 183, 0, 0},
           {14139, 2862, 937, 0, 0},    {23404, 12044, 6453, 0, 0},
           {28127, 20450, 14674, 0, 0}, {30010, 25381, 21189, 0, 0},
           {7335, 926, 299, 0, 0},      {13973, 3479, 1357, 0, 0},
           {25124, 15184, 9176, 0, 0},  {29360, 23754, 17721, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}}},
         {{{28232, 22696, 18767, 0, 0}, {7309, 1352, 562, 0, 0},
           {16163, 4720, 1950, 0, 0},   {21760, 9911, 5049, 0, 0},
           {25853, 16500, 10453, 0, 0}, {30143, 25956, 22231, 0, 0},
           {8511, 980, 269, 0, 0},      {15888, 3314, 889, 0, 0},
           {20810, 7714, 2990, 0, 0},   {24852, 14050, 7684, 0, 0},
           {29385, 23991, 19322, 0, 0}, {10048, 1165, 375, 0, 0},
           {17808, 4643, 1433, 0, 0},   {23037, 10558, 4840, 0, 0},
           {26464, 16936, 10491, 0, 0}, {29858, 24950, 20602, 0, 0},
           {12393, 2141, 637, 0, 0},    {18864, 5484, 1881, 0, 0},
           {23400, 11210, 5624, 0, 0},  {26831, 17802, 11649, 0, 0},
           {30101, 25543, 21449, 0, 0}, {8798, 1298, 390, 0, 0},
           {15595, 3034, 750, 0, 0},    {19973, 7327, 2803, 0, 0},
           {23787, 13088, 6875, 0, 0},  {28040, 21396, 15866, 0, 0},
           {8481, 971, 329, 0, 0},      {16065, 3623, 1072, 0, 0},
           {21935, 9214, 4043, 0, 0},   {26300, 16202, 9711, 0, 0},
           {30353, 26206, 22490, 0, 0}, {6158, 373, 109, 0, 0},
           {14178, 2270, 651, 0, 0},    {20348, 7012, 2818, 0, 0},
           {25129, 14022, 8058, 0, 0},  {29767, 24682, 20421, 0, 0},
           {7692, 704, 188, 0, 0},      {14822, 2640, 740, 0, 0},
           {20744, 7783, 3390, 0, 0},   {25251, 14378, 8464, 0, 0},
           {29525, 23987, 19437, 0, 0}, {24576, 16384, 8192, 0, 0}},
          {{26731, 15997, 10811, 0, 0}, {7994, 1064, 342, 0, 0},
           {15938, 4179, 1712, 0, 0},   {22166, 9940, 5008, 0, 0},
           {26035, 15939, 9697, 0, 0},  {29518, 23854, 19212, 0, 0},
           {7186, 548, 100, 0, 0},      {14109, 2426, 545, 0, 0},
           {20222, 6619, 2253, 0, 0},   {24348, 12317, 5967, 0, 0},
           {28132, 20348, 14424, 0, 0}, {5187, 406, 129, 0, 0},
           {13781, 2685, 790, 0, 0},    {21441, 8520, 3684, 0, 0},
           {25504, 15049, 8648, 0, 0},  {28773, 22000, 16599, 0, 0},
           {6875, 937, 281, 0, 0},      {16191, 4181, 1389, 0, 0},
           {22579, 10020, 4586, 0, 0},  {25936, 15674, 9212, 0, 0},
           {29060, 22658, 17434, 0, 0}, {6864, 486, 112, 0, 0},
           {13047, 1976, 492, 0, 0},    {19949, 6525, 2357, 0, 0},
           {24196, 12154, 5877, 0, 0},  {27404, 18709, 12301, 0, 0},
           {6188, 330, 91, 0, 0},       {11916, 1543, 428, 0, 0},
           {20333, 7068, 2801, 0, 0},   {24077, 11943, 5792, 0, 0},
           {28322, 20559, 15499, 0, 0}, {5418, 339, 72, 0, 0},
           {11396, 1791, 496, 0, 0},    {20095, 7498, 2915, 0, 0},
           {23560, 11843, 6128, 0, 0},  {27750, 19417, 14036, 0, 0},
           {5417, 289, 55, 0, 0},       {11370, 1559, 381, 0, 0},
           {20606, 7721, 2926, 0, 0},   {24872, 14077, 7449, 0, 0},
           {28098, 19886, 13887, 0, 0}, {24576, 16384, 8192, 0, 0}}},
         {{{27281, 22308, 19060, 0, 0}, {11171, 4465, 2094, 0, 0},
           {21731, 10815, 6292, 0, 0},  {24621, 14806, 9816, 0, 0},
           {27526, 19707, 14236, 0, 0}, {30879, 27560, 24586, 0, 0},
           {5994, 635, 178, 0, 0},      {14924, 3204, 1001, 0, 0},
           {21078, 8330, 3597, 0, 0},   {25226, 14553, 8309, 0, 0},
           {29775, 24718, 20449, 0, 0}, {4745, 440, 177, 0, 0},
           {14117, 2642, 814, 0, 0},    {20604, 7622, 3179, 0, 0},
           {25006, 14238, 7997, 0, 0},  {29276, 23585, 18848, 0, 0},
           {5177, 760, 277, 0, 0},      {15619, 3915, 1258, 0, 0},
           {21283, 8765, 3908, 0, 0},   {25071, 14682, 8558, 0, 0},
           {29693, 24769, 20550, 0, 0}, {4500, 286, 114, 0, 0},
           {13137, 1717, 364, 0, 0},    {18908, 5508, 1748, 0, 0},
           {23163, 11155, 5174, 0, 0},  {27892, 20606, 14860, 0, 0},
           {5520, 452, 192, 0, 0},      {13813, 2311, 693, 0, 0},
           {20944, 8771, 3973, 0, 0},   {25422, 14572, 8121, 0, 0},
           {29365, 23521, 18657, 0, 0}, {3057, 113, 33, 0, 0},
           {11599, 1374, 351, 0, 0},    {19281, 5570, 1811, 0, 0},
           {23940, 11085, 5154, 0, 0},  {28498, 21317, 15730, 0, 0},
           {4060, 190, 37, 0, 0},       {12648, 1527, 286, 0, 0},
           {19076, 5218, 1447, 0, 0},   {23350, 10254, 4329, 0, 0},
           {27769, 19485, 13306, 0, 0}, {24576, 16384, 8192, 0, 0}},
          {{27095, 18466, 13057, 0, 0}, {6517, 2067, 934, 0, 0},
           {19986, 8985, 4965, 0, 0},   {23641, 12111, 6960, 0, 0},
           {26400, 16560, 11306, 0, 0}, {30303, 25591, 21946, 0, 0},
           {2807, 205, 49, 0, 0},       {14450, 2877, 819, 0, 0},
           {21407, 8254, 3411, 0, 0},   {24868, 13165, 7161, 0, 0},
           {28766, 22178, 17222, 0, 0}, {3131, 458, 173, 0, 0},
           {14472, 2855, 959, 0, 0},    {22624, 11253, 5897, 0, 0},
           {27410, 18446, 12374, 0, 0}, {29701, 24406, 19422, 0, 0},
           {4116, 298, 92, 0, 0},       {15230, 1997, 559, 0, 0},
           {18844, 5886, 2274, 0, 0},   {22272, 9931, 4899, 0, 0},
           {25532, 16372, 11147, 0, 0}, {2025, 81, 22, 0, 0},
           {9762, 1092, 279, 0, 0},     {18274, 4940, 1648, 0, 0},
           {22594, 9967, 4416, 0, 0},   {26526, 17487, 11725, 0, 0},
           {6951, 525, 48, 0, 0},       {14150, 1401, 443, 0, 0},
           {18771, 4450, 890, 0, 0},    {20513, 6234, 1385, 0, 0},
           {23207, 11180, 4318, 0, 0},  {4580, 133, 44, 0, 0},
           {10708, 403, 40, 0, 0},      {14666, 2078, 240, 0, 0},
           {18572, 3904, 769, 0, 0},    {20506, 6976, 1903, 0, 0},
           {8592, 659, 140, 0, 0},      {14488, 3087, 805, 0, 0},
           {22563, 9065, 3104, 0, 0},   {24879, 12743, 5092, 0, 0},
           {26708, 16025, 8798, 0, 0},  {24576, 16384, 8192, 0, 0}}},
         {{{27627, 25672, 24508, 0, 0}, {5582, 3746, 2979, 0, 0},
           {26100, 20200, 17086, 0, 0}, {30596, 26587, 24130, 0, 0},
           {31642, 29389, 28237, 0, 0}, {32325, 31407, 30514, 0, 0},
           {6685, 1615, 332, 0, 0},     {19282, 8165, 4285, 0, 0},
           {26260, 17928, 12858, 0, 0}, {29382, 23968, 19482, 0, 0},
           {31238, 28446, 25714, 0, 0}, {3129, 688, 220, 0, 0},
           {16871, 5216, 2478, 0, 0},   {24180, 12721, 7385, 0, 0},
           {27879, 19429, 13499, 0, 0}, {30528, 25897, 22270, 0, 0},
           {4603, 571, 251, 0, 0},      {12033, 2341, 1200, 0, 0},
           {18443, 8097, 5076, 0, 0},   {27649, 20214, 14963, 0, 0},
           {30958, 27327, 24507, 0, 0}, {1556, 44, 20, 0, 0},
           {9416, 1002, 223, 0, 0},     {18099, 5198, 1709, 0, 0},
           {24276, 11874, 5496, 0, 0},  {29124, 22574, 17564, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}},
          {{30307, 25755, 23397, 0, 0}, {8019, 3168, 1782, 0, 0},
           {23302, 13731, 10351, 0, 0}, {29184, 23488, 18368, 0, 0},
           {31263, 28839, 27335, 0, 0}, {32091, 31268, 30032, 0, 0},
           {8781, 2066, 651, 0, 0},     {19214, 8197, 3505, 0, 0},
           {26557, 18212, 11613, 0, 0}, {29633, 21796, 17143, 0, 0},
           {30333, 25641, 21341, 0, 0}, {1468, 236, 218, 0, 0},
           {18011, 2403, 814, 0, 0},    {28363, 21156, 14215, 0, 0},
           {32188, 28636, 25446, 0, 0}, {31073, 22599, 18644, 0, 0},
           {2760, 486, 177, 0, 0},      {13524, 2660, 1020, 0, 0},
           {21588, 8610, 3213, 0, 0},   {27118, 17796, 13559, 0, 0},
           {30654, 27659, 24312, 0, 0}, {912, 52, 20, 0, 0},
           {9756, 1104, 196, 0, 0},     {19074, 6112, 2132, 0, 0},
           {24626, 13260, 6675, 0, 0},  {28515, 21813, 16044, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}}},
         {{{32167, 31785, 31457, 0, 0}, {14043, 9362, 4681, 0, 0},
           {27307, 24576, 21845, 0, 0}, {28987, 17644, 11343, 0, 0},
           {30181, 25007, 20696, 0, 0}, {32662, 32310, 31958, 0, 0},
           {10486, 3058, 874, 0, 0},    {24260, 11842, 6784, 0, 0},
           {29042, 20055, 14685, 0, 0}, {31148, 25656, 21875, 0, 0},
           {32039, 30532, 29273, 0, 0}, {2605, 294, 84, 0, 0},
           {14464, 2304, 768, 0, 0},    {21325, 6242, 3121, 0, 0},
           {26761, 17476, 11469, 0, 0}, {30534, 26065, 23831, 0, 0},
           {1814, 591, 197, 0, 0},      {15405, 3206, 1692, 0, 0},
           {23082, 10304, 5358, 0, 0},  {24576, 16384, 11378, 0, 0},
           {31013, 24722, 21504, 0, 0}, {1600, 34, 20, 0, 0},
           {10282, 1327, 297, 0, 0},    {19935, 7141, 3030, 0, 0},
           {25788, 15389, 9646, 0, 0},  {29657, 23881, 19289, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}},
          {{24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0}}}},
        {{{{26727, 20914, 16841, 0, 0}, {12442, 1863, 517, 0, 0},
           {18604, 5937, 2043, 0, 0},   {23008, 12121, 6183, 0, 0},
           {26352, 17815, 11549, 0, 0}, {29802, 25617, 21877, 0, 0},
           {9201, 1394, 514, 0, 0},     {17790, 5352, 1822, 0, 0},
           {23334, 12543, 6514, 0, 0},  {26110, 18210, 12233, 0, 0},
           {28852, 24091, 19779, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {14680, 3223, 1181, 0, 0},
           {19706, 6925, 2695, 0, 0},   {23828, 15941, 10517, 0, 0},
           {25114, 19548, 14795, 0, 0}, {27035, 22452, 18312, 0, 0},
           {9889, 1380, 654, 0, 0},     {17553, 4775, 1813, 0, 0},
           {23371, 13323, 7790, 0, 0},  {29326, 22955, 17424, 0, 0},
           {31400, 28832, 26236, 0, 0}, {7274, 735, 362, 0, 0},
           {15996, 4805, 2050, 0, 0},   {23349, 14603, 9508, 0, 0},
           {30091, 25267, 20971, 0, 0}, {31252, 28424, 25598, 0, 0},
           {6212, 1314, 667, 0, 0},     {15640, 5733, 2660, 0, 0},
           {24444, 17424, 12519, 0, 0}, {30865, 27072, 23299, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}},
          {{24313, 13765, 8400, 0, 0},  {9205, 747, 164, 0, 0},
           {16531, 3322, 833, 0, 0},    {22044, 8769, 3410, 0, 0},
           {26043, 15240, 8352, 0, 0},  {28841, 21841, 15943, 0, 0},
           {6455, 480, 134, 0, 0},      {15338, 2673, 673, 0, 0},
           {21652, 8162, 3089, 0, 0},   {25573, 14384, 7499, 0, 0},
           {28042, 19916, 13453, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {9946, 1120, 285, 0, 0},
           {16044, 3135, 839, 0, 0},    {22507, 9735, 4043, 0, 0},
           {25739, 14928, 8240, 0, 0},  {27901, 18882, 11266, 0, 0},
           {7470, 876, 277, 0, 0},      {14959, 3438, 1256, 0, 0},
           {23100, 11439, 6189, 0, 0},  {27994, 19812, 13792, 0, 0},
           {30446, 25738, 21228, 0, 0}, {7296, 848, 225, 0, 0},
           {14811, 3381, 1136, 0, 0},   {23572, 12175, 6368, 0, 0},
           {28088, 20063, 13566, 0, 0}, {29851, 24312, 19332, 0, 0},
           {6297, 709, 194, 0, 0},      {14310, 2985, 859, 0, 0},
           {24368, 13304, 6812, 0, 0},  {28956, 21795, 15562, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}}},
         {{{25989, 19025, 15090, 0, 0}, {7962, 971, 311, 0, 0},
           {15152, 3721, 1396, 0, 0},   {21705, 9593, 4765, 0, 0},
           {26247, 16658, 10444, 0, 0}, {30004, 25264, 21114, 0, 0},
           {7502, 401, 131, 0, 0},      {13714, 2215, 593, 0, 0},
           {20629, 7556, 2961, 0, 0},   {25457, 14606, 8064, 0, 0},
           {29371, 23604, 18694, 0, 0}, {6780, 560, 246, 0, 0},
           {16515, 3856, 1242, 0, 0},   {23617, 11381, 5396, 0, 0},
           {27080, 17853, 11272, 0, 0}, {30051, 25141, 20764, 0, 0},
           {9624, 913, 325, 0, 0},      {16698, 4277, 1443, 0, 0},
           {24066, 12301, 6251, 0, 0},  {27525, 18812, 12401, 0, 0},
           {30147, 25433, 21201, 0, 0}, {6132, 428, 138, 0, 0},
           {12778, 1718, 427, 0, 0},    {19525, 6663, 2453, 0, 0},
           {24180, 13247, 6850, 0, 0},  {28051, 21183, 15464, 0, 0},
           {6924, 476, 186, 0, 0},      {13678, 2133, 671, 0, 0},
           {20805, 8222, 3829, 0, 0},   {26550, 16681, 10414, 0, 0},
           {30428, 26160, 22342, 0, 0}, {4722, 192, 74, 0, 0},
           {11590, 1455, 472, 0, 0},    {19282, 6584, 2898, 0, 0},
           {25619, 14897, 9045, 0, 0},  {29935, 24810, 20509, 0, 0},
           {5058, 240, 82, 0, 0},       {12094, 1692, 500, 0, 0},
           {20355, 7813, 3525, 0, 0},   {26092, 15841, 9671, 0, 0},
           {29802, 24435, 19849, 0, 0}, {24576, 16384, 8192, 0, 0}},
          {{24129, 13429, 8339, 0, 0},  {8364, 931, 243, 0, 0},
           {15771, 3343, 984, 0, 0},    {21515, 8534, 3619, 0, 0},
           {26017, 15374, 8740, 0, 0},  {29278, 22938, 17577, 0, 0},
           {6485, 297, 54, 0, 0},       {13169, 1600, 326, 0, 0},
           {19622, 5814, 1875, 0, 0},   {24554, 12180, 5878, 0, 0},
           {28069, 19687, 13468, 0, 0}, {4556, 310, 99, 0, 0},
           {14174, 2452, 668, 0, 0},    {21549, 8360, 3534, 0, 0},
           {25903, 15112, 8619, 0, 0},  {29090, 22406, 16762, 0, 0},
           {6943, 632, 152, 0, 0},      {15455, 2915, 747, 0, 0},
           {21571, 8297, 3296, 0, 0},   {25821, 14987, 8363, 0, 0},
           {29000, 22108, 16507, 0, 0}, {5416, 268, 62, 0, 0},
           {11918, 1300, 299, 0, 0},    {18747, 5061, 1635, 0, 0},
           {23804, 11020, 4930, 0, 0},  {27331, 18103, 11581, 0, 0},
           {6464, 276, 70, 0, 0},       {12359, 1388, 383, 0, 0},
           {19086, 5546, 2136, 0, 0},   {23794, 11532, 6083, 0, 0},
           {28534, 21103, 15834, 0, 0}, {6495, 411, 57, 0, 0},
           {12096, 1526, 327, 0, 0},    {18596, 5514, 1866, 0, 0},
           {22898, 10870, 5493, 0, 0},  {27604, 19262, 13498, 0, 0},
           {6043, 309, 40, 0, 0},       {11777, 1326, 241, 0, 0},
           {19697, 6334, 1957, 0, 0},   {24584, 12678, 6026, 0, 0},
           {27965, 19513, 12873, 0, 0}, {24576, 16384, 8192, 0, 0}}},
         {{{25213, 17826, 14267, 0, 0}, {8358, 1590, 481, 0, 0},
           {18374, 6030, 2515, 0, 0},   {24355, 13214, 7573, 0, 0},
           {28002, 19844, 13983, 0, 0}, {30739, 26962, 23561, 0, 0},
           {5992, 404, 105, 0, 0},      {14036, 2801, 837, 0, 0},
           {21763, 8982, 3916, 0, 0},   {26302, 15859, 9258, 0, 0},
           {29724, 24130, 19349, 0, 0}, {3560, 186, 64, 0, 0},
           {12700, 1911, 560, 0, 0},    {20765, 7683, 3173, 0, 0},
           {25821, 15018, 8579, 0, 0},  {29523, 23665, 18761, 0, 0},
           {5409, 303, 99, 0, 0},       {13347, 2154, 594, 0, 0},
           {20853, 7758, 3189, 0, 0},   {25818, 15092, 8694, 0, 0},
           {29761, 24295, 19672, 0, 0}, {3766, 92, 33, 0, 0},
           {10666, 919, 192, 0, 0},     {18360, 4759, 1363, 0, 0},
           {23741, 11089, 4837, 0, 0},  {28074, 20090, 14020, 0, 0},
           {4552, 240, 86, 0, 0},       {11919, 1504, 450, 0, 0},
           {20012, 6953, 3017, 0, 0},   {25203, 13967, 7845, 0, 0},
           {29259, 23235, 18291, 0, 0}, {2635, 81, 29, 0, 0},
           {9705, 858, 253, 0, 0},      {18180, 4717, 1636, 0, 0},
           {23683, 11119, 5311, 0, 0},  {28507, 21114, 15504, 0, 0},
           {3250, 77, 20, 0, 0},        {10317, 809, 155, 0, 0},
           {17904, 4046, 1068, 0, 0},   {23073, 9804, 4052, 0, 0},
           {27836, 19410, 13266, 0, 0}, {24576, 16384, 8192, 0, 0}},
          {{26303, 15810, 11080, 0, 0}, {7569, 1254, 408, 0, 0},
           {17994, 5619, 2161, 0, 0},   {23511, 11330, 5796, 0, 0},
           {27045, 17585, 10886, 0, 0}, {29618, 23889, 19037, 0, 0},
           {5779, 506, 86, 0, 0},       {15372, 2831, 683, 0, 0},
           {21381, 7867, 2984, 0, 0},   {25479, 13947, 7220, 0, 0},
           {29034, 22191, 16682, 0, 0}, {3040, 267, 73, 0, 0},
           {15337, 3067, 865, 0, 0},    {22847, 9942, 4468, 0, 0},
           {26872, 17334, 10700, 0, 0}, {29338, 23122, 18011, 0, 0},
           {4154, 257, 63, 0, 0},       {13404, 2130, 505, 0, 0},
           {19639, 6514, 2366, 0, 0},   {24014, 12284, 6328, 0, 0},
           {28390, 21161, 15658, 0, 0}, {2476, 97, 24, 0, 0},
           {10988, 1165, 267, 0, 0},    {18454, 4939, 1477, 0, 0},
           {23157, 10441, 4505, 0, 0},  {27878, 19681, 13703, 0, 0},
           {6906, 201, 35, 0, 0},       {11974, 718, 201, 0, 0},
           {15525, 2143, 514, 0, 0},    {19485, 5140, 1294, 0, 0},
           {23099, 10236, 3850, 0, 0},  {5333, 71, 20, 0, 0},
           {7846, 378, 54, 0, 0},       {11319, 1264, 232, 0, 0},
           {16376, 3039, 936, 0, 0},    {21076, 7884, 3692, 0, 0},
           {8575, 478, 33, 0, 0},       {13859, 1664, 205, 0, 0},
           {20532, 5927, 1365, 0, 0},   {24597, 10928, 3686, 0, 0},
           {25544, 15488, 7493, 0, 0},  {24576, 16384, 8192, 0, 0}}},
         {{{29690, 25929, 22878, 0, 0}, {18931, 12318, 8289, 0, 0},
           {26854, 18546, 13440, 0, 0}, {28902, 22501, 18006, 0, 0},
           {30156, 25560, 21726, 0, 0}, {31701, 29777, 27992, 0, 0},
           {6951, 1122, 239, 0, 0},     {19060, 6430, 2383, 0, 0},
           {25440, 14183, 7898, 0, 0},  {28077, 19688, 13492, 0, 0},
           {30943, 27515, 24416, 0, 0}, {3382, 453, 144, 0, 0},
           {15608, 3767, 1408, 0, 0},   {23166, 10906, 5372, 0, 0},
           {26853, 16996, 10620, 0, 0}, {29982, 24989, 20721, 0, 0},
           {3522, 318, 105, 0, 0},      {14072, 2839, 950, 0, 0},
           {22258, 9399, 4208, 0, 0},   {26539, 16269, 9643, 0, 0},
           {30160, 25320, 21063, 0, 0}, {2015, 58, 20, 0, 0},
           {11130, 1281, 265, 0, 0},    {19831, 5914, 1898, 0, 0},
           {24586, 12172, 5798, 0, 0},  {29131, 22499, 17271, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}},
          {{27524, 20618, 15862, 0, 0}, {12282, 5910, 3067, 0, 0},
           {25012, 14451, 9033, 0, 0},  {29316, 23512, 19622, 0, 0},
           {30748, 27562, 24539, 0, 0}, {30967, 27775, 24865, 0, 0},
           {5717, 910, 237, 0, 0},      {16780, 5237, 2149, 0, 0},
           {23580, 11284, 6049, 0, 0},  {26495, 15582, 8968, 0, 0},
           {29660, 23413, 18004, 0, 0}, {1692, 248, 88, 0, 0},
           {14649, 2731, 918, 0, 0},    {22524, 9799, 5296, 0, 0},
           {28076, 18691, 13495, 0, 0}, {29074, 21091, 15212, 0, 0},
           {2708, 187, 48, 0, 0},       {11757, 1993, 648, 0, 0},
           {20837, 7948, 3479, 0, 0},   {25649, 15106, 8412, 0, 0},
           {28935, 22062, 16464, 0, 0}, {814, 37, 20, 0, 0},
           {8855, 1044, 279, 0, 0},     {17248, 4708, 1482, 0, 0},
           {21251, 9760, 4197, 0, 0},   {26575, 18260, 12139, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}}},
         {{{31733, 29961, 28612, 0, 0}, {19606, 14630, 11829, 0, 0},
           {30072, 26135, 24013, 0, 0}, {31395, 28607, 25915, 0, 0},
           {31669, 30022, 28052, 0, 0}, {32428, 31747, 31169, 0, 0},
           {9942, 2349, 633, 0, 0},     {22373, 11006, 5826, 0, 0},
           {28042, 20361, 15407, 0, 0}, {30321, 25688, 22175, 0, 0},
           {31541, 29051, 26757, 0, 0}, {4612, 1344, 834, 0, 0},
           {15853, 5014, 2395, 0, 0},   {23620, 11778, 6337, 0, 0},
           {26818, 17253, 11620, 0, 0}, {30276, 25441, 21242, 0, 0},
           {2166, 291, 98, 0, 0},       {12742, 2813, 1200, 0, 0},
           {21548, 9140, 4663, 0, 0},   {26116, 15749, 9795, 0, 0},
           {29704, 24232, 19725, 0, 0}, {999, 44, 20, 0, 0},
           {10538, 1881, 395, 0, 0},    {20534, 7689, 3037, 0, 0},
           {25442, 13952, 7415, 0, 0},  {28835, 21861, 16152, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}},
          {{24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0}}}},
        {{{{23872, 16541, 12138, 0, 0}, {9139, 986, 241, 0, 0},
           {17595, 5013, 1447, 0, 0},   {22610, 11535, 5386, 0, 0},
           {26348, 17911, 11210, 0, 0}, {29499, 24613, 20122, 0, 0},
           {7933, 759, 272, 0, 0},      {16259, 4347, 1189, 0, 0},
           {21811, 11254, 5350, 0, 0},  {24887, 16838, 10672, 0, 0},
           {27380, 21808, 16850, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {12023, 1995, 675, 0, 0},
           {17568, 5547, 1907, 0, 0},   {19736, 11895, 7101, 0, 0},
           {20483, 14105, 9274, 0, 0},  {21205, 15287, 11279, 0, 0},
           {6508, 786, 448, 0, 0},      {17371, 4685, 1668, 0, 0},
           {23026, 13551, 7944, 0, 0},  {29507, 23139, 17406, 0, 0},
           {31288, 28446, 25269, 0, 0}, {5169, 512, 308, 0, 0},
           {15911, 5109, 1994, 0, 0},   {23217, 14478, 9020, 0, 0},
           {29716, 23835, 18665, 0, 0}, {30747, 26858, 22981, 0, 0},
           {3763, 753, 376, 0, 0},      {15091, 5074, 1905, 0, 0},
           {23564, 15412, 9549, 0, 0},  {30365, 25252, 19954, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}},
          {{21960, 10712, 5872, 0, 0},  {7029, 455, 92, 0, 0},
           {15480, 2565, 547, 0, 0},    {21409, 7890, 2872, 0, 0},
           {25819, 15001, 7875, 0, 0},  {28481, 20972, 14697, 0, 0},
           {4888, 247, 63, 0, 0},       {13730, 1764, 354, 0, 0},
           {20204, 6423, 2000, 0, 0},   {24499, 12821, 5989, 0, 0},
           {27094, 18111, 11094, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {7026, 449, 97, 0, 0},
           {13211, 1604, 314, 0, 0},    {19387, 6387, 2013, 0, 0},
           {22667, 11302, 6046, 0, 0},  {23559, 13118, 5943, 0, 0},
           {5661, 851, 336, 0, 0},      {14712, 3875, 1565, 0, 0},
           {22568, 11334, 6004, 0, 0},  {28108, 19855, 13266, 0, 0},
           {30400, 25838, 20264, 0, 0}, {5808, 610, 155, 0, 0},
           {14140, 2763, 737, 0, 0},    {22535, 10326, 4536, 0, 0},
           {27297, 18138, 11252, 0, 0}, {29533, 22001, 15659, 0, 0},
           {5072, 328, 76, 0, 0},       {12736, 1601, 330, 0, 0},
           {24068, 11427, 4326, 0, 0},  {27106, 17937, 10973, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}}},
         {{{23064, 15474, 11636, 0, 0}, {6006, 490, 135, 0, 0},
           {14386, 3148, 949, 0, 0},    {21877, 9293, 4045, 0, 0},
           {26410, 16185, 9459, 0, 0},  {29520, 23650, 18627, 0, 0},
           {5564, 195, 69, 0, 0},       {12950, 1944, 439, 0, 0},
           {20996, 7648, 2727, 0, 0},   {25773, 14735, 7729, 0, 0},
           {29016, 22326, 16670, 0, 0}, {5546, 512, 209, 0, 0},
           {17412, 4369, 1293, 0, 0},   {23947, 12133, 5711, 0, 0},
           {27257, 18364, 11529, 0, 0}, {29833, 24546, 19717, 0, 0},
           {7893, 648, 239, 0, 0},      {17535, 4503, 1323, 0, 0},
           {24163, 12198, 5836, 0, 0},  {27337, 18355, 11572, 0, 0},
           {29774, 24427, 19545, 0, 0}, {4567, 164, 68, 0, 0},
           {11727, 1322, 312, 0, 0},    {19547, 6555, 2293, 0, 0},
           {24513, 13383, 6731, 0, 0},  {27838, 20183, 13938, 0, 0},
           {4000, 320, 141, 0, 0},      {13063, 2207, 747, 0, 0},
           {21196, 9179, 4548, 0, 0},   {27236, 17734, 11322, 0, 0},
           {30308, 25618, 21312, 0, 0}, {2894, 149, 69, 0, 0},
           {11147, 1697, 567, 0, 0},    {20257, 8021, 3776, 0, 0},
           {26487, 16373, 10020, 0, 0}, {29522, 23490, 18271, 0, 0},
           {3053, 143, 56, 0, 0},       {11810, 1757, 485, 0, 0},
           {21535, 9097, 3962, 0, 0},   {26756, 16640, 9900, 0, 0},
           {29341, 22917, 17354, 0, 0}, {24576, 16384, 8192, 0, 0}},
          {{21752, 10657, 5974, 0, 0},  {6822, 411, 91, 0, 0},
           {14878, 2316, 516, 0, 0},    {21090, 7626, 2952, 0, 0},
           {26048, 15234, 8184, 0, 0},  {28538, 21103, 14948, 0, 0},
           {4368, 145, 21, 0, 0},       {11604, 1100, 193, 0, 0},
           {19196, 5380, 1586, 0, 0},   {24534, 12018, 5410, 0, 0},
           {27703, 18713, 11871, 0, 0}, {3787, 221, 63, 0, 0},
           {14087, 2225, 529, 0, 0},    {21849, 8693, 3482, 0, 0},
           {26337, 15569, 8691, 0, 0},  {28949, 22304, 16150, 0, 0},
           {5898, 301, 75, 0, 0},       {13727, 1937, 421, 0, 0},
           {20974, 7557, 2752, 0, 0},   {25880, 14749, 7798, 0, 0},
           {28398, 20405, 13776, 0, 0}, {3190, 98, 24, 0, 0},
           {9609, 761, 155, 0, 0},      {17453, 4099, 1092, 0, 0},
           {23470, 10161, 3986, 0, 0},  {26624, 16855, 9800, 0, 0},
           {4658, 269, 99, 0, 0},       {11194, 1831, 753, 0, 0},
           {20009, 7950, 4041, 0, 0},   {26223, 16007, 9726, 0, 0},
           {29119, 22171, 15935, 0, 0}, {4605, 216, 40, 0, 0},
           {10667, 1299, 304, 0, 0},    {19608, 7296, 2625, 0, 0},
           {25465, 14084, 7300, 0, 0},  {27527, 18793, 11813, 0, 0},
           {4368, 137, 24, 0, 0},       {10664, 975, 165, 0, 0},
           {19211, 6197, 1922, 0, 0},   {25019, 12907, 6093, 0, 0},
           {27895, 18738, 11534, 0, 0}, {24576, 16384, 8192, 0, 0}}},
         {{{22968, 15133, 11695, 0, 0}, {6615, 883, 241, 0, 0},
           {17730, 4916, 1762, 0, 0},   {24050, 12204, 6282, 0, 0},
           {27640, 18692, 12254, 0, 0}, {30132, 25202, 20843, 0, 0},
           {5217, 264, 67, 0, 0},       {14458, 2714, 668, 0, 0},
           {22557, 9348, 3686, 0, 0},   {26546, 15892, 8852, 0, 0},
           {29306, 22814, 17270, 0, 0}, {2777, 135, 47, 0, 0},
           {12885, 2017, 567, 0, 0},    {21627, 8584, 3483, 0, 0},
           {26348, 15828, 8994, 0, 0},  {29376, 23015, 17650, 0, 0},
           {4303, 152, 56, 0, 0},       {12918, 2066, 524, 0, 0},
           {21785, 8744, 3545, 0, 0},   {26474, 15998, 9186, 0, 0},
           {29524, 23485, 18259, 0, 0}, {2745, 51, 20, 0, 0},
           {9828, 736, 142, 0, 0},      {18486, 4840, 1295, 0, 0},
           {24206, 11441, 4854, 0, 0},  {27922, 19375, 12849, 0, 0},
           {2787, 178, 73, 0, 0},       {12303, 1805, 602, 0, 0},
           {21289, 9189, 4573, 0, 0},   {26852, 17120, 10695, 0, 0},
           {29737, 24163, 19370, 0, 0}, {1622, 77, 29, 0, 0},
           {9662, 1044, 324, 0, 0},     {18985, 6030, 2329, 0, 0},
           {24916, 13300, 6961, 0, 0},  {28908, 21644, 15915, 0, 0},
           {1754, 44, 20, 0, 0},        {9139, 659, 140, 0, 0},
           {18021, 4653, 1365, 0, 0},   {24223, 11526, 5290, 0, 0},
           {28194, 19987, 13701, 0, 0}, {24576, 16384, 8192, 0, 0}},
          {{23583, 13074, 8080, 0, 0},  {6687, 783, 147, 0, 0},
           {16753, 3768, 981, 0, 0},    {22226, 9078, 3562, 0, 0},
           {26036, 14823, 8091, 0, 0},  {28852, 21729, 16046, 0, 0},
           {4544, 202, 24, 0, 0},       {13668, 1630, 283, 0, 0},
           {20240, 6148, 1889, 0, 0},   {25027, 12491, 5883, 0, 0},
           {28202, 19923, 13778, 0, 0}, {2835, 175, 50, 0, 0},
           {15098, 2435, 613, 0, 0},    {22383, 9168, 3859, 0, 0},
           {26525, 16532, 10361, 0, 0}, {28792, 22379, 16751, 0, 0},
           {4391, 207, 30, 0, 0},       {13402, 1593, 286, 0, 0},
           {19441, 5593, 1674, 0, 0},   {24510, 11999, 5625, 0, 0},
           {28065, 19570, 13241, 0, 0}, {1682, 62, 20, 0, 0},
           {9915, 866, 185, 0, 0},      {18009, 4582, 1349, 0, 0},
           {23484, 10386, 4420, 0, 0},  {27183, 17576, 10900, 0, 0},
           {4477, 116, 22, 0, 0},       {12919, 661, 197, 0, 0},
           {17934, 5950, 3554, 0, 0},   {22462, 10174, 4096, 0, 0},
           {26153, 15384, 9384, 0, 0},  {3821, 164, 23, 0, 0},
           {7143, 479, 122, 0, 0},      {14010, 4096, 1365, 0, 0},
           {22751, 9338, 4245, 0, 0},   {25906, 17499, 10637, 0, 0},
           {8835, 259, 29, 0, 0},       {12841, 1273, 137, 0, 0},
           {20865, 6745, 2147, 0, 0},   {25742, 12674, 5516, 0, 0},
           {26770, 14662, 8331, 0, 0},  {24576, 16384, 8192, 0, 0}}},
         {{{28312, 21494, 17235, 0, 0}, {11549, 3689, 1152, 0, 0},
           {21595, 8994, 4201, 0, 0},   {25486, 14475, 8505, 0, 0},
           {27878, 19482, 13653, 0, 0}, {30878, 27260, 24109, 0, 0},
           {6117, 632, 121, 0, 0},      {18138, 4514, 1313, 0, 0},
           {24052, 11481, 5373, 0, 0},  {27153, 17437, 10760, 0, 0},
           {30093, 25068, 20618, 0, 0}, {2814, 242, 78, 0, 0},
           {16642, 3786, 1135, 0, 0},   {23738, 11407, 5416, 0, 0},
           {27357, 17975, 11497, 0, 0}, {29825, 24346, 19605, 0, 0},
           {3229, 167, 38, 0, 0},       {14643, 2383, 567, 0, 0},
           {22346, 8678, 3300, 0, 0},   {26300, 15281, 8330, 0, 0},
           {29798, 24115, 19237, 0, 0}, {1856, 53, 20, 0, 0},
           {12102, 1395, 271, 0, 0},    {20259, 6128, 1851, 0, 0},
           {24710, 12139, 5478, 0, 0},  {28537, 20762, 14716, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}},
          {{22566, 12135, 7284, 0, 0},  {5432, 1323, 416, 0, 0},
           {20348, 8384, 4216, 0, 0},   {25120, 14653, 8912, 0, 0},
           {27106, 18427, 12866, 0, 0}, {29157, 22440, 17378, 0, 0},
           {1823, 152, 32, 0, 0},       {14086, 2263, 515, 0, 0},
           {21255, 7432, 2565, 0, 0},   {25319, 13316, 6620, 0, 0},
           {28286, 19717, 13882, 0, 0}, {746, 78, 21, 0, 0},
           {14190, 2267, 622, 0, 0},    {21519, 9400, 4137, 0, 0},
           {27123, 15810, 10610, 0, 0}, {27759, 21324, 16131, 0, 0},
           {1411, 58, 20, 0, 0},        {11216, 1274, 264, 0, 0},
           {18877, 5091, 1428, 0, 0},   {23717, 10670, 4596, 0, 0},
           {27578, 19391, 13282, 0, 0}, {404, 28, 20, 0, 0},
           {7929, 861, 217, 0, 0},      {15608, 3989, 1072, 0, 0},
           {20316, 8631, 3166, 0, 0},   {26603, 17379, 10291, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}}},
         {{{30193, 25487, 21691, 0, 0}, {18766, 11902, 7366, 0, 0},
           {26425, 17712, 13110, 0, 0}, {28294, 20910, 15727, 0, 0},
           {29903, 24469, 20234, 0, 0}, {31424, 28819, 26377, 0, 0},
           {8048, 1529, 309, 0, 0},     {20183, 7412, 2800, 0, 0},
           {25587, 14522, 8324, 0, 0},  {27743, 19101, 12883, 0, 0},
           {30247, 25464, 21163, 0, 0}, {2860, 516, 184, 0, 0},
           {15347, 3612, 1193, 0, 0},   {22879, 10580, 4986, 0, 0},
           {26890, 17121, 10645, 0, 0}, {29954, 24103, 19445, 0, 0},
           {2585, 200, 55, 0, 0},       {14240, 2573, 719, 0, 0},
           {21786, 8162, 3111, 0, 0},   {25811, 14603, 7537, 0, 0},
           {29260, 22650, 17300, 0, 0}, {1007, 32, 20, 0, 0},
           {11727, 1440, 222, 0, 0},    {20200, 6036, 1602, 0, 0},
           {24716, 12048, 5035, 0, 0},  {28432, 20576, 14372, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}},
          {{24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0}}}},
        {{{{25706, 16296, 10449, 0, 0}, {8230, 507, 94, 0, 0},
           {19093, 4727, 989, 0, 0},    {24178, 12094, 5137, 0, 0},
           {27083, 18093, 10755, 0, 0}, {29113, 22870, 17037, 0, 0},
           {6275, 350, 110, 0, 0},      {16392, 3426, 678, 0, 0},
           {22174, 10119, 3798, 0, 0},  {24592, 15598, 8465, 0, 0},
           {27163, 20074, 13629, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {8880, 866, 226, 0, 0},
           {14156, 3081, 781, 0, 0},    {16523, 7916, 3519, 0, 0},
           {17003, 10160, 5209, 0, 0},  {12873, 8069, 5258, 0, 0},
           {4367, 556, 311, 0, 0},      {17494, 4943, 1788, 0, 0},
           {23404, 14640, 8436, 0, 0},  {30485, 24575, 17686, 0, 0},
           {31540, 28796, 24887, 0, 0}, {3313, 299, 148, 0, 0},
           {14787, 4523, 1380, 0, 0},   {21847, 12670, 6528, 0, 0},
           {29025, 20939, 14111, 0, 0}, {30394, 23175, 17053, 0, 0},
           {1700, 302, 133, 0, 0},      {12447, 3196, 797, 0, 0},
           {21997, 12513, 5649, 0, 0},  {29973, 22358, 15407, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}},
          {{23448, 10666, 4928, 0, 0},  {5711, 304, 44, 0, 0},
           {16437, 2500, 459, 0, 0},    {22449, 8833, 3048, 0, 0},
           {26579, 16320, 8662, 0, 0},  {29179, 21884, 13960, 0, 0},
           {3742, 144, 20, 0, 0},       {13542, 1261, 181, 0, 0},
           {20076, 5847, 1565, 0, 0},   {25719, 13236, 5133, 0, 0},
           {25041, 17099, 9516, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {4712, 143, 20, 0, 0},
           {10385, 693, 99, 0, 0},      {17351, 5670, 1019, 0, 0},
           {14641, 6275, 5578, 0, 0},   {27307, 16384, 10923, 0, 0},
           {4786, 677, 184, 0, 0},      {13723, 2900, 796, 0, 0},
           {22371, 10502, 4836, 0, 0},  {26778, 19071, 11268, 0, 0},
           {30976, 25856, 17664, 0, 0}, {4570, 267, 50, 0, 0},
           {11234, 1247, 199, 0, 0},    {21659, 7551, 2751, 0, 0},
           {27097, 17644, 6617, 0, 0},  {28087, 18725, 14043, 0, 0},
           {4080, 188, 27, 0, 0},       {10192, 689, 107, 0, 0},
           {22141, 10627, 4428, 0, 0},  {23406, 18725, 4681, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}}},
         {{{25014, 15820, 10626, 0, 0}, {7098, 438, 77, 0, 0},
           {17105, 3543, 774, 0, 0},    {22890, 9480, 3610, 0, 0},
           {26349, 15680, 8432, 0, 0},  {28909, 21765, 15729, 0, 0},
           {5206, 173, 43, 0, 0},       {15193, 2180, 369, 0, 0},
           {21949, 7930, 2459, 0, 0},   {25644, 14082, 6852, 0, 0},
           {28289, 20080, 13428, 0, 0}, {4383, 292, 95, 0, 0},
           {17462, 3763, 830, 0, 0},    {23831, 11153, 4446, 0, 0},
           {26786, 17165, 9982, 0, 0},  {29148, 22501, 16632, 0, 0},
           {5488, 304, 101, 0, 0},      {17161, 3608, 764, 0, 0},
           {23677, 10633, 4028, 0, 0},  {26536, 16136, 8748, 0, 0},
           {28721, 21391, 15096, 0, 0}, {3548, 138, 50, 0, 0},
           {13118, 1548, 306, 0, 0},    {19718, 6456, 1941, 0, 0},
           {23540, 11898, 5300, 0, 0},  {26622, 17619, 10797, 0, 0},
           {2599, 287, 145, 0, 0},      {15556, 3457, 1214, 0, 0},
           {22857, 11457, 5886, 0, 0},  {28281, 19454, 12396, 0, 0},
           {30198, 24996, 19879, 0, 0}, {1844, 155, 60, 0, 0},
           {13278, 2562, 661, 0, 0},    {21536, 8770, 3492, 0, 0},
           {25999, 14813, 7733, 0, 0},  {28370, 20145, 13554, 0, 0},
           {2159, 141, 46, 0, 0},       {13398, 2186, 481, 0, 0},
           {22311, 9149, 3359, 0, 0},   {26325, 15131, 7934, 0, 0},
           {28123, 19532, 12662, 0, 0}, {24576, 16384, 8192, 0, 0}},
          {{24142, 12497, 6552, 0, 0},  {6061, 362, 57, 0, 0},
           {15769, 2439, 482, 0, 0},    {21323, 7645, 2482, 0, 0},
           {26357, 13940, 7167, 0, 0},  {25967, 20310, 12520, 0, 0},
           {2850, 86, 20, 0, 0},        {12119, 1029, 150, 0, 0},
           {19889, 4995, 1187, 0, 0},   {24872, 11017, 4524, 0, 0},
           {27508, 17898, 9070, 0, 0},  {3516, 175, 37, 0, 0},
           {15696, 2308, 474, 0, 0},    {22115, 8625, 3403, 0, 0},
           {26232, 15278, 8785, 0, 0},  {27839, 19598, 12683, 0, 0},
           {4631, 250, 53, 0, 0},       {14597, 1984, 361, 0, 0},
           {21331, 7332, 2309, 0, 0},   {25516, 14234, 6592, 0, 0},
           {28642, 19415, 11790, 0, 0}, {1606, 42, 20, 0, 0},
           {9751, 546, 67, 0, 0},       {17139, 3535, 722, 0, 0},
           {23381, 10147, 3288, 0, 0},  {25846, 15152, 7758, 0, 0},
           {3930, 503, 154, 0, 0},      {13067, 2562, 848, 0, 0},
           {21554, 10358, 4835, 0, 0},  {27448, 18591, 9734, 0, 0},
           {27719, 19887, 14941, 0, 0}, {5284, 297, 34, 0, 0},
           {11692, 1242, 207, 0, 0},    {20061, 6465, 1557, 0, 0},
           {24599, 11046, 4549, 0, 0},  {26723, 13362, 5726, 0, 0},
           {5015, 196, 23, 0, 0},       {11936, 890, 115, 0, 0},
           {19518, 5412, 1094, 0, 0},   {25050, 11260, 2910, 0, 0},
           {25559, 14418, 7209, 0, 0},  {24576, 16384, 8192, 0, 0}}},
         {{{24892, 15867, 11027, 0, 0}, {8767, 870, 143, 0, 0},
           {18239, 4809, 1317, 0, 0},   {24495, 11950, 5510, 0, 0},
           {27490, 18095, 11258, 0, 0}, {29785, 23925, 18729, 0, 0},
           {4752, 194, 36, 0, 0},       {15297, 2462, 467, 0, 0},
           {22544, 8705, 3040, 0, 0},   {26166, 14814, 7716, 0, 0},
           {28766, 21183, 15009, 0, 0}, {2578, 134, 29, 0, 0},
           {15271, 2486, 498, 0, 0},    {22539, 9039, 3230, 0, 0},
           {26424, 15557, 8328, 0, 0},  {28919, 21579, 15660, 0, 0},
           {4198, 185, 42, 0, 0},       {15247, 2607, 530, 0, 0},
           {22615, 9203, 3390, 0, 0},   {26313, 15427, 8325, 0, 0},
           {28861, 21726, 15744, 0, 0}, {2079, 53, 20, 0, 0},
           {11222, 928, 158, 0, 0},     {19221, 5187, 1309, 0, 0},
           {23856, 11011, 4459, 0, 0},  {27220, 17688, 10722, 0, 0},
           {1985, 228, 83, 0, 0},       {15228, 3240, 1100, 0, 0},
           {22608, 11300, 5985, 0, 0},  {28044, 19375, 12714, 0, 0},
           {30066, 24594, 19666, 0, 0}, {1120, 82, 26, 0, 0},
           {11814, 1674, 431, 0, 0},    {20348, 7070, 2589, 0, 0},
           {25464, 13448, 6520, 0, 0},  {28402, 20507, 13904, 0, 0},
           {1187, 45, 20, 0, 0},        {11395, 1182, 243, 0, 0},
           {20024, 6143, 1883, 0, 0},   {25337, 12446, 5818, 0, 0},
           {28076, 19445, 12657, 0, 0}, {24576, 16384, 8192, 0, 0}},
          {{24935, 14399, 8673, 0, 0},  {6118, 495, 66, 0, 0},
           {16397, 2807, 577, 0, 0},    {21713, 8686, 3139, 0, 0},
           {25876, 14124, 7368, 0, 0},  {27762, 19711, 13528, 0, 0},
           {2934, 102, 20, 0, 0},       {13191, 1433, 198, 0, 0},
           {20515, 6259, 1646, 0, 0},   {24777, 11996, 5057, 0, 0},
           {27091, 16858, 9709, 0, 0},  {2659, 236, 48, 0, 0},
           {16021, 2602, 516, 0, 0},    {22634, 9226, 3584, 0, 0},
           {26977, 16592, 9212, 0, 0},  {28406, 22354, 15484, 0, 0},
           {3276, 142, 20, 0, 0},       {12874, 1366, 243, 0, 0},
           {19826, 5697, 1899, 0, 0},   {24422, 11552, 5363, 0, 0},
           {26196, 15681, 8909, 0, 0},  {733, 33, 20, 0, 0},
           {9811, 930, 150, 0, 0},      {18044, 4196, 996, 0, 0},
           {22404, 8769, 3215, 0, 0},   {25764, 14335, 7113, 0, 0},
           {5240, 491, 87, 0, 0},       {15809, 1597, 672, 0, 0},
           {22282, 9175, 4806, 0, 0},   {24576, 16384, 9557, 0, 0},
           {23831, 14895, 11916, 0, 0}, {5053, 766, 153, 0, 0},
           {17695, 3277, 1092, 0, 0},   {21504, 8192, 4096, 0, 0},
           {30427, 14043, 9362, 0, 0},  {25486, 14564, 7282, 0, 0},
           {4221, 555, 111, 0, 0},      {11980, 2995, 529, 0, 0},
           {25988, 11299, 2260, 0, 0},  {26810, 17873, 8937, 0, 0},
           {16384, 10923, 5461, 0, 0},  {24576, 16384, 8192, 0, 0}}},
         {{{26776, 18464, 13003, 0, 0}, {10156, 1530, 312, 0, 0},
           {19312, 5606, 1681, 0, 0},   {24767, 12706, 6264, 0, 0},
           {27600, 18663, 12004, 0, 0}, {30136, 24997, 20383, 0, 0},
           {5734, 424, 59, 0, 0},       {16918, 3353, 771, 0, 0},
           {23274, 9992, 3927, 0, 0},   {26617, 15938, 8799, 0, 0},
           {29307, 22729, 17046, 0, 0}, {2634, 199, 37, 0, 0},
           {17130, 3346, 823, 0, 0},    {23618, 10903, 4550, 0, 0},
           {27121, 17049, 10092, 0, 0}, {29366, 22996, 17291, 0, 0},
           {4238, 182, 33, 0, 0},       {15629, 2470, 476, 0, 0},
           {22568, 8729, 3083, 0, 0},   {26349, 15094, 7982, 0, 0},
           {29224, 22543, 16944, 0, 0}, {1435, 42, 20, 0, 0},
           {12150, 1281, 224, 0, 0},    {19867, 5551, 1536, 0, 0},
           {24144, 11034, 4597, 0, 0},  {27664, 18577, 12020, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}},
          {{21562, 11678, 6207, 0, 0},  {4009, 489, 97, 0, 0},
           {18597, 4816, 1199, 0, 0},   {23025, 9861, 3627, 0, 0},
           {25897, 14882, 7900, 0, 0},  {27808, 19616, 13453, 0, 0},
           {1691, 107, 20, 0, 0},       {13368, 1573, 253, 0, 0},
           {20016, 5910, 1728, 0, 0},   {24398, 10670, 4177, 0, 0},
           {27311, 17395, 10470, 0, 0}, {1071, 62, 20, 0, 0},
           {14908, 2111, 435, 0, 0},    {20258, 7956, 3507, 0, 0},
           {26588, 13644, 8046, 0, 0},  {27727, 19220, 14809, 0, 0},
           {1216, 52, 20, 0, 0},        {10860, 999, 145, 0, 0},
           {18298, 4567, 1203, 0, 0},   {23275, 9786, 4160, 0, 0},
           {25910, 15528, 8631, 0, 0},  {225, 16, 12, 0, 0},
           {8482, 671, 102, 0, 0},      {16810, 3551, 744, 0, 0},
           {22561, 8534, 2810, 0, 0},   {25839, 14463, 7116, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}}},
         {{{28631, 21921, 17086, 0, 0}, {14944, 5767, 2710, 0, 0},
           {22564, 9972, 4477, 0, 0},   {26692, 16833, 10643, 0, 0},
           {28916, 21831, 15952, 0, 0}, {30516, 26444, 22637, 0, 0},
           {6928, 752, 106, 0, 0},      {17659, 4500, 1237, 0, 0},
           {23383, 10537, 4428, 0, 0},  {26686, 16096, 9289, 0, 0},
           {29450, 23341, 18087, 0, 0}, {2174, 194, 50, 0, 0},
           {15932, 3216, 909, 0, 0},    {23212, 10226, 4412, 0, 0},
           {26463, 16043, 9228, 0, 0},  {29392, 22873, 17584, 0, 0},
           {3385, 151, 23, 0, 0},       {13877, 1959, 367, 0, 0},
           {21080, 6826, 2081, 0, 0},   {25300, 13299, 6117, 0, 0},
           {28859, 21410, 15756, 0, 0}, {1204, 32, 20, 0, 0},
           {11862, 1157, 168, 0, 0},    {19577, 5147, 1231, 0, 0},
           {24000, 10739, 4092, 0, 0},  {27689, 18659, 11862, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0},  {24576, 16384, 8192, 0, 0}},
          {{24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0},
           {24576, 16384, 8192, 0, 0}, {24576, 16384, 8192, 0, 0}}}}};

constexpr uint16_t kDefaultCoeffBaseRangeCdf[kNumPlaneTypes]
    [kCoeffBaseRangeContexts][kCoeffBaseRangeSymbolCount] = {
        {{{{18470, 12050, 8594},  {20232, 13167, 8979},
           {24056, 17717, 13265}, {26598, 21441, 17334},
           {28026, 23842, 20230}, {28965, 25451, 22222},
           {31072, 29451, 27897}, {18376, 12817, 10012},
           {16790, 9550, 5950},   {20581, 13294, 8879},
           {23592, 17128, 12509}, {25700, 20113, 15740},
           {27112, 22326, 18296}, {30188, 27776, 25524},
           {20632, 14719, 11342}, {18984, 12047, 8287},
           {21932, 15147, 10868}, {24396, 18324, 13921},
           {26245, 20989, 16768}, {27431, 22870, 19008},
           {29734, 26908, 24306}},
          {{16801, 9863, 6482},   {19234, 12114, 8189},
           {23264, 16676, 12233}, {25793, 20200, 15865},
           {27404, 22677, 18748}, {28411, 24398, 20911},
           {30262, 27834, 25550}, {9736, 3953, 1832},
           {13228, 6064, 3049},   {17610, 9799, 5671},
           {21360, 13903, 9118},  {23883, 17320, 12518},
           {25660, 19915, 15352}, {28537, 24727, 21288},
           {12945, 6278, 3612},   {13878, 6839, 3836},
           {17108, 9277, 5335},   {20621, 12992, 8280},
           {23040, 15994, 11119}, {24849, 18491, 13702},
           {27328, 22598, 18583}}},
         {{{18362, 11906, 8354},  {20944, 13861, 9659},
           {24511, 18375, 13965}, {26908, 22021, 17990},
           {28293, 24282, 20784}, {29162, 25814, 22725},
           {31032, 29358, 27720}, {18338, 12722, 9886},
           {17175, 9869, 6059},   {20666, 13400, 8957},
           {23709, 17184, 12506}, {25769, 20165, 15720},
           {27084, 22271, 18215}, {29946, 27330, 24906},
           {16983, 11183, 8409},  {14421, 7539, 4502},
           {17794, 10281, 6379},  {21345, 14087, 9497},
           {23905, 17418, 12760}, {25615, 19916, 15490},
           {29061, 25732, 22786}},
          {{17308, 11072, 7299},  {20598, 13519, 9577},
           {24045, 17741, 13436}, {26340, 21064, 16894},
           {27846, 23476, 19716}, {28629, 25073, 21758},
           {30477, 28260, 26170}, {12912, 5848, 2940},
           {14845, 7479, 3976},   {18490, 10800, 6471},
           {21858, 14632, 9818},  {24345, 17953, 13141},
           {25997, 20485, 15994}, {28694, 25018, 21687},
           {12916, 6694, 4096},   {13397, 6658, 3779},
           {16503, 8895, 5105},   {20010, 12390, 7816},
           {22673, 15670, 10807}, {24518, 18140, 13317},
           {27563, 23023, 19146}}},
         {{{22205, 16535, 13005}, {22974, 16746, 12964},
           {26018, 20823, 17009}, {27805, 23582, 20016},
           {28923, 25333, 22141}, {29717, 26683, 23934},
           {31457, 30172, 28938}, {21522, 16364, 13079},
           {20453, 13857, 10037}, {22211, 15673, 11479},
           {24632, 18762, 14519}, {26420, 21294, 17203},
           {27572, 23113, 19368}, {30419, 28242, 26181},
           {19431, 14038, 11199}, {13462, 6697, 3886},
           {16816, 9228, 5514},   {20359, 12834, 8338},
           {23008, 16062, 11379}, {24764, 18548, 13950},
           {28630, 24974, 21807}},
          {{21898, 16084, 11819}, {23104, 17538, 14088},
           {25882, 20659, 17360}, {27943, 23868, 20463},
           {29138, 25606, 22454}, {29732, 26339, 23381},
           {31097, 29472, 27828}, {18949, 13609, 9742},
           {20784, 13660, 9648},  {22078, 15558, 11105},
           {24784, 18614, 14435}, {25900, 20474, 16644},
           {27494, 23774, 19900}, {29780, 26997, 24344},
           {13032, 6121, 3627},   {13835, 6698, 3784},
           {16989, 9720, 5568},   {20130, 12707, 8236},
           {22076, 15223, 10548}, {23551, 17517, 12714},
           {27690, 23484, 20174}}},
         {{{30437, 29106, 27524}, {29877, 27997, 26623},
           {28170, 25145, 23039}, {29248, 25923, 23569},
           {29351, 26649, 23444}, {30167, 27356, 25383},
           {32168, 31595, 31024}, {25096, 19482, 15299},
           {28536, 24976, 21975}, {29853, 27451, 25371},
           {30450, 28412, 26616}, {30641, 28768, 27214},
           {30918, 29290, 27493}, {31791, 30835, 29925},
           {14488, 8381, 4779},   {16916, 10097, 6583},
           {18923, 11817, 7979},  {21713, 14802, 10639},
           {23630, 17346, 12967}, {25314, 19623, 15312},
           {29398, 26375, 23755}},
          {{26926, 23539, 21930}, {30455, 29277, 28492},
           {29770, 26664, 25272}, {30348, 25321, 22900},
           {29734, 24273, 21845}, {28692, 23831, 21793},
           {31682, 30398, 29469}, {23054, 15514, 12324},
           {24225, 19070, 15645}, {27850, 23761, 20858},
           {28639, 25236, 22215}, {30404, 27235, 24710},
           {30934, 29222, 27205}, {31295, 29860, 28635},
           {17363, 11575, 7149},  {17077, 10816, 6207},
           {19806, 13574, 8603},  {22496, 14913, 10639},
           {24180, 17498, 12050}, {24086, 18099, 13268},
           {27898, 23132, 19563}}},
         {{{24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}},
          {{24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}, {24576, 16384, 8192},
           {24576, 16384, 8192}}}},

};

/* clang-format on */

}  // namespace libgav1

// Normalizes a CDF in bounds tab_size so that it fits in APROBA_MAX.
static void ANSNormalizeCDF(uint32_t max_symbol, uint32_t tab_size,
                            uint16_t* const cdf) {
  assert(max_symbol > 0 && max_symbol <= APROBA_MAX_SYMBOL);
  assert(tab_size > 0);
  assert(cdf[0] == 0);
  uint32_t counts[APROBA_MAX_SYMBOL];
  for (uint32_t i = 0; i < max_symbol - 1; ++i) {
    assert(cdf[i + 1] >= cdf[i] && cdf[i + 1] <= tab_size);
    counts[i] = cdf[i + 1] - cdf[i];
  }
  counts[max_symbol - 1] = tab_size - cdf[max_symbol - 1];

  ANSNormalizeCounts(counts, max_symbol, APROBA_MAX);

  cdf[0] = 0;
  for (uint32_t i = 1; i < max_symbol; ++i) {
    cdf[i] = cdf[i - 1] + (uint16_t)counts[i - 1];
  }
}

// Prints a C array.
template <class T>
void PrintNormalized(const T& cdf_array) {
  uint32_t c[kCoordMaxSize + 1];
  uint32_t num_extents;
  CoordinateFiller<kCoordMaxSize>::Do(cdf_array, c, &num_extents);
  // We subtract an extra 1 because we always use 0 as the first coordinates.
  std::cout << std::string(num_extents - 1 - 1, '{');
  uint32_t num_clusters = c[0];
  for (uint32_t i = 1; i < num_extents - 1; ++i) num_clusters *= c[i];
  // Remove the last element that is just a counter.
  const uint32_t cdf_size = c[num_extents - 1] - 1;
  // We divide by c[0] because we always use 0 as the first coordinates.
  num_clusters /= c[0];
  for (uint32_t i = 0; i < num_clusters; ++i) {
    const uint16_t* const cdf = GetCDF(i, cdf_array);
    uint16_t cdf_wp2[APROBA_MAX_SYMBOL];
    std::copy(cdf, cdf + cdf_size, cdf_wp2);
    std::reverse(cdf_wp2, cdf_wp2 + cdf_size);
    ANSNormalizeCDF(cdf_size, /*tab_size=*/1 << 15, cdf_wp2);
    std::cout << "{";
    for (uint32_t i = 0; i < cdf_size; ++i) {
      std::cout << cdf_wp2[i];
      if (i + 1 < cdf_size) std::cout << ", ";
    }
    std::cout << "}";
    uint32_t num_p = 0;
    for (int j = num_extents - 2; j >= 0; --j) {
      uint32_t prod = 1;
      for (uint32_t k = j; k <= num_extents - 2; ++k) prod *= c[k];
      // Check how many brackets we close.
      if ((i + 1) % prod == 0) {
        ++num_p;
      } else {
        break;
      }
    }
    if (num_p > 0) {
      std::cout << std::string(num_p, '}');
      if (i != num_clusters - 1) {
        std::cout << ", " << std::endl << std::string(num_p, '{');
      }
    } else {
      std::cout << ", ";
    }
  }
  std::cout << ";" << std::endl << std::endl;
}

}  // namespace WP2

// Main file that prints the different AOM CDFs after WP2 normalization.
int main() {
  std::cout << "constexpr uint16_t kDefaultEobPt4Cdf[kNumPlaneTypes]"
               "[kEobPtContexts][kEobPt4SymbolCount] =";
  WP2::PrintNormalized(WP2::libgav1::kDefaultEobPt4Cdf);
  std::cout << "constexpr uint16_t kDefaultEobPt8Cdf[kNumPlaneTypes]"
               "[kEobPtContexts][kEobPt8SymbolCount] =";
  WP2::PrintNormalized(WP2::libgav1::kDefaultEobPt8Cdf);
  std::cout << "constexpr uint16_t kDefaultEobPt16Cdf[kNumPlaneTypes]"
               "[kEobPtContexts][kEobPt16SymbolCount] =";
  WP2::PrintNormalized(WP2::libgav1::kDefaultEobPt16Cdf);
  std::cout << "constexpr uint16_t kDefaultEobPt32Cdf[kNumPlaneTypes]"
               "[kEobPtContexts][kEobPt32SymbolCount] =";
  WP2::PrintNormalized(WP2::libgav1::kDefaultEobPt32Cdf);
  std::cout << "constexpr uint16_t kDefaultEobPt64Cdf[kNumPlaneTypes]"
               "[kEobPtContexts][kEobPt64SymbolCount] =";
  WP2::PrintNormalized(WP2::libgav1::kDefaultEobPt64Cdf);
  std::cout << "constexpr uint16_t kDefaultEobPt128Cdf[kNumPlaneTypes]"
               "[kEobPtContexts][kEobPt128SymbolCount] =";
  WP2::PrintNormalized(WP2::libgav1::kDefaultEobPt128Cdf);
  std::cout << "constexpr uint16_t kDefaultEobPt256Cdf[kNumPlaneTypes]"
               "[kEobPtContexts][kEobPt256SymbolCount] =";
  WP2::PrintNormalized(WP2::libgav1::kDefaultEobPt256Cdf);
  std::cout << "constexpr uint16_t kDefaultEobPt512Cdf[kNumPlaneTypes]"
               "[kEobPt512SymbolCount] =";
  WP2::PrintNormalized(WP2::libgav1::kDefaultEobPt512Cdf);
  std::cout << "constexpr uint16_t kDefaultEobPt1024Cdf[kNumPlaneTypes]"
               "[kEobPt1024SymbolCount] =";
  WP2::PrintNormalized(WP2::libgav1::kDefaultEobPt1024Cdf);
  std::cout << "constexpr uint16_t kDefaultEobExtraCdf["
               "kNumPlaneTypes][kEobExtraContexts]["
               "kBooleanFieldCdfSize - 1] =";
  WP2::PrintNormalized(WP2::libgav1::kDefaultEobExtraCdf);
  std::cout << "constexpr uint16_t kDefaultCoeffBaseEobCdf["
               "kNumPlaneTypes]["
               "kCoeffBaseEobContexts][kCoeffBaseEobSymbolCount] =";
  WP2::PrintNormalized(WP2::libgav1::kDefaultCoeffBaseEobCdf);
  std::cout << "constexpr uint16_t kDefaultCoeffBaseCdf["
               "kNumPlaneTypes][kCoeffBaseContexts]["
               "kCoeffBaseSymbolCount] =";
  WP2::PrintNormalized(WP2::libgav1::kDefaultCoeffBaseCdf);
  std::cout << "constexpr uint16_t kDefaultCoeffBaseRangeCdf["
               "kNumPlaneTypes]["
               "kCoeffBaseRangeContexts][kCoeffBaseRangeSymbolCount + 1] =";
  WP2::PrintNormalized(WP2::libgav1::kDefaultCoeffBaseRangeCdf);
  return 0;
}

namespace WP2 {
namespace libgav1 {

#else


constexpr uint16_t kDefaultEobPt4Cdf[kNumPlaneTypes][kEobPtContexts]
                                    [kEobPt4SymbolCount] = {
                                        {{0, 13936, 15394}, {0, 14148, 15442}},
                                        {{0, 9102, 11540}, {0, 9060, 12502}}};

constexpr uint16_t
    kDefaultEobPt8Cdf[kNumPlaneTypes][kEobPtContexts][kEobPt8SymbolCount] = {
        {{0, 13936, 15393, 15864}, {0, 14148, 15442, 16048}},
        {{0, 9102, 11540, 13909}, {0, 9060, 12502, 14707}}};

constexpr uint16_t
    kDefaultEobPt16Cdf[kNumPlaneTypes][kEobPtContexts][kEobPt16SymbolCount] = {
        {{0, 13936, 15393, 15864, 15964}, {0, 14148, 15442, 16048, 16199}},
        {{0, 9102, 11539, 13908, 14760}, {0, 9060, 12502, 14707, 15432}}};

constexpr uint16_t
    kDefaultEobPt32Cdf[kNumPlaneTypes][kEobPtContexts][kEobPt32SymbolCount] = {
        {{0, 13112, 15332, 15895, 16124, 16184},
         {0, 12615, 14720, 15726, 16181, 16279}},
        {{0, 6183, 10486, 12589, 14247, 15066},
         {0, 6856, 10704, 12932, 14794, 15491}}};

constexpr uint16_t
    kDefaultEobPt64Cdf[kNumPlaneTypes][kEobPtContexts][kEobPt64SymbolCount] = {
        {{0, 12504, 14750, 15490, 15832, 16134, 16219},
         {0, 9885, 12005, 13636, 15653, 16018, 16216}},
        {{0, 4698, 7541, 9476, 11340, 13731, 14631},
         {0, 5143, 9029, 10927, 13945, 15033, 15602}}};

constexpr uint16_t
    kDefaultEobPt128Cdf[kNumPlaneTypes][kEobPtContexts][kEobPt128SymbolCount] =
        {{{0, 10090, 13368, 14542, 15337, 15813, 16142, 16274},
          {0, 7635, 10102, 11644, 13968, 15756, 16034, 16198}},
         {{0, 2580, 4418, 6368, 8459, 9945, 12656, 13762},
          {0, 3008, 6089, 7263, 9268, 13449, 14648, 15357}}};

constexpr uint16_t
    kDefaultEobPt256Cdf[kNumPlaneTypes][kEobPtContexts][kEobPt256SymbolCount] =
        {{{0, 8557, 10706, 12077, 13299, 14589, 15440, 16092, 16229},
          {0, 3592, 4934, 6440, 7714, 13582, 14885, 15459, 15885}},
         {{0, 4299, 6406, 7604, 10095, 11949, 13408, 14764, 15124},
          {0, 2753, 3579, 4543, 6058, 11014, 12666, 14318, 15282}}};

constexpr uint16_t kDefaultEobPt512Cdf[kNumPlaneTypes][kEobPt512SymbolCount] = {
    {0, 3353, 4678, 6990, 8905, 11267, 13668, 14530, 15892, 16063},
    {0, 1813, 3319, 5923, 7390, 8375, 9707, 11386, 13161, 13837}};

constexpr uint16_t kDefaultEobPt1024Cdf[kNumPlaneTypes][kEobPt1024SymbolCount] =
    {{0, 3468, 5098, 7360, 9711, 13207, 14803, 15572, 16008, 16173, 16187},
     {0, 492, 2111, 2757, 5707, 8115, 11117, 14262, 14918, 15389, 15451}};

constexpr uint16_t kDefaultEobExtraCdf[kNumPlaneTypes]
    [kEobExtraContexts][kBooleanFieldCdfSize - 1] = {
       {{0, 7904},
        {0, 7773},
        {0, 12573},
        {0, 8192},
        {0, 8192},
        {0, 8192},
        {0, 8192},
        {0, 8192},
        {0, 8192}},
       {{0, 6850},
        {0, 5122},
        {0, 9695},
        {0, 8192},
        {0, 8192},
        {0, 8192},
        {0, 8192},
        {0, 8192},
        {0, 8192}}};

constexpr uint16_t kDefaultCoeffBaseEobCdf
    [kNumPlaneTypes][kCoeffBaseEobContexts][kCoeffBaseEobSymbolCount] = {
        {{0, 1857, 7466}, {0, 661, 1584}, {0, 445, 962}, {0, 1910, 3921}},
        {{0, 1371, 5702}, {0, 173, 1129}, {0, 74, 556}, {0, 444, 1570}}};

constexpr uint16_t kDefaultCoeffBaseCdf[kNumPlaneTypes][kCoeffBaseContexts]
    [kCoeffBaseSymbolCount] = {
         {{0, 10020, 11919, 14367}, {0, 446, 1514, 7344},
          {0, 1138, 3323, 10086},   {0, 2882, 5826, 11661},
          {0, 5744, 8813, 13230},   {0, 11492, 13205, 15152},
          {0, 420, 1112, 6051},     {0, 1317, 3323, 9862},
          {0, 3995, 7006, 12308},   {0, 7067, 9964, 13756},
          {0, 10803, 12781, 14974}, {0, 4096, 8192, 12288},
          {0, 4096, 8192, 12288},   {0, 4096, 8192, 12288},
          {0, 4096, 8192, 12288},   {0, 4096, 8192, 12288},
          {0, 4096, 8192, 12288},   {0, 4096, 8192, 12288},
          {0, 4096, 8192, 12288},   {0, 4096, 8192, 12288},
          {0, 4096, 8192, 12288},   {0, 1082, 2608, 8517},
          {0, 1991, 4488, 10779},   {0, 6533, 9447, 13410},
          {0, 9438, 11722, 14276},  {0, 11354, 13153, 14967},
          {0, 593, 1188, 6582},     {0, 1260, 3258, 9623},
          {0, 4128, 7073, 12161},   {0, 8396, 11212, 14475},
          {0, 12985, 14325, 15643}, {0, 289, 733, 5060},
          {0, 1160, 2821, 8970},    {0, 4731, 7532, 12227},
          {0, 9990, 12233, 14873},  {0, 12792, 14178, 15616},
          {0, 698, 1497, 5207},     {0, 1777, 3648, 9023},
          {0, 7052, 9511, 13047},   {0, 11723, 13501, 15350},
          {0, 4096, 8192, 12288},   {0, 4096, 8192, 12288}},
         {{0, 5503, 8162, 13233},   {0, 147, 616, 4865},
          {0, 636, 2158, 8786},     {0, 2127, 4911, 11374},
          {0, 4633, 7953, 13117},   {0, 8846, 11476, 14615},
          {0, 122, 447, 4163},      {0, 567, 1923, 8444},
          {0, 2151, 4948, 11423},   {0, 4538, 7901, 13120},
          {0, 7773, 10732, 14327},  {0, 4096, 8192, 12288},
          {0, 4096, 8192, 12288},   {0, 4096, 8192, 12288},
          {0, 4096, 8192, 12288},   {0, 4096, 8192, 12288},
          {0, 4096, 8192, 12288},   {0, 4096, 8192, 12288},
          {0, 4096, 8192, 12288},   {0, 4096, 8192, 12288},
          {0, 4096, 8192, 12288},   {0, 280, 1000, 6284},
          {0, 755, 2349, 9008},     {0, 3126, 6292, 12195},
          {0, 5394, 8734, 13426},   {0, 7433, 10592, 14250},
          {0, 94, 372, 4204},       {0, 413, 1332, 7048},
          {0, 2596, 5262, 11286},   {0, 6175, 9209, 13636},
          {0, 10634, 12676, 15046}, {0, 92, 406, 4046},
          {0, 469, 1432, 7070},     {0, 3227, 6023, 11702},
          {0, 7336, 10224, 14063},  {0, 10594, 12690, 15005},
          {0, 150, 464, 3668},      {0, 679, 1740, 6987},
          {0, 4588, 7592, 12562},   {0, 8860, 11877, 14680},
          {0, 4096, 8192, 12288},   {0, 4096, 8192, 12288}}
};

constexpr uint16_t kDefaultCoeffBaseRangeCdf[kNumPlaneTypes]
    [kCoeffBaseRangeContexts][kCoeffBaseRangeSymbolCount + 1] = {
         {{0, 4297, 6025, 9235},    {0, 4490, 6584, 10117},
          {0, 6632, 8858, 12028},   {0, 8666, 10720, 13299},
          {0, 10115, 11921, 14013}, {0, 11110, 12725, 14482},
          {0, 13948, 14725, 15536}, {0, 5006, 6409, 9189},
          {0, 2975, 4775, 8395},    {0, 4439, 6647, 10291},
          {0, 6254, 8564, 11796},   {0, 7869, 10056, 12850},
          {0, 9148, 11163, 13556},  {0, 12762, 13888, 15094},
          {0, 5671, 7360, 10317},   {0, 4144, 6024, 9493},
          {0, 5433, 7573, 10966},   {0, 6960, 9162, 12198},
          {0, 8383, 10494, 13122},  {0, 9503, 11434, 13715},
          {0, 12153, 13454, 14867}},
         {{0, 3241, 4932, 8401},    {0, 4095, 6058, 9618},
          {0, 6116, 8338, 11632},   {0, 7932, 10100, 12897},
          {0, 9373, 11338, 13702},  {0, 10455, 12199, 14206},
          {0, 12775, 13917, 15131}, {0, 916, 1977, 4869},
          {0, 1525, 3033, 6615},    {0, 2836, 4900, 8806},
          {0, 4559, 6952, 10681},   {0, 6258, 8659, 11941},
          {0, 7675, 9957, 12830},   {0, 10643, 12363, 14268},
          {0, 1806, 3139, 6473},    {0, 1918, 3420, 6940},
          {0, 2668, 4639, 8555},    {0, 4140, 6496, 10311},
          {0, 5559, 7997, 11520},   {0, 6850, 9245, 12424},
          {0, 9291, 11299, 13664}}
};

#endif

}  // namespace libgav1
}  // namespace WP2
